import os
import typing as t
from datetime import datetime

import pandas as pd
import requests
from loguru import logger


class LinearConnector:

    url = "https://api.linear.app/graphql"

    def __init__(self, access_token: str = None) -> None:

        if access_token is None:
            access_token = os.getenv("LINEAR_ACCESS_TOKEN")

        if access_token is None:
            raise Exception("No Linear access token provided")

        self.access_token = access_token

        self._check_permissions()

    def _post_request(self, query, variables={}):
        r = requests.post(
            self.url,
            json={"query": query, "variables": variables},
            headers={"Authorization": self.access_token},
        )
        if r.status_code == 200:
            return r.status_code, r.json()
        else:
            raise Exception(
                f"LinearConnector initialization failed: {r.status_code} {r.text}"
            )

    def _check_permissions(self):
        query = """
            query Me {
                viewer {
                    id
                    name
                    email
                }
            }"""
        return self._post_request(query)

    def get_team_and_projects(self, team_name: str) -> t.Tuple[str, str, dict]:
        query = """
            query ProjectsByTeam($team_name: String){
                teams(filter: {name: {eq: $team_name}}) {
                    nodes {
                        id
                        name
                        projects{
                            nodes{
                                id
                                name
                                startDate
                                targetDate
                                state
                                url
                            }
                        }
                    }
                }
            }
            """
        variables = {"team_name": team_name}
        _, response = self._post_request(query, variables)
        response = response["data"]["teams"]["nodes"]
        assert len(response) == 1, f"We expect only one team with the name {team_name}"

        team_id = response[0]["id"]
        team_name = response[0]["name"]
        projects = response[0]["projects"]["nodes"]

        projects = [
            {
                "linear_project_id": p["id"],
                "linear_project_name": p["name"],
                "linear_start_date": p["startDate"],
                "linear_target_date": p["targetDate"],
                "linear_state": p["state"],
                "linear_url": p["url"],
            }
            for p in projects
        ]

        if len(projects) == 0:
            columns = [
                "linear_project_id",
                "linear_project_name",
                "linear_start_date",
                "linear_target_date",
                "linear_state",
                "linear_url",
            ]
            projects = pd.DataFrame(columns=columns)
        else:
            projects = pd.DataFrame(projects)

        return team_id, team_name, projects

    def create_project(self, name: str, team_id: str) -> str:
        query = """
            query ProjectsByName($project_name: String!) {
                projects(filter: {name: {eq: $project_name}}) {
                    nodes {
                        id
                        teams {
                            nodes {
                                name
                                id
                            }
                        }
                    }
                }
            }
        """
        variables = {"project_name": name}
        _, response = self._post_request(query, variables)
        project_ids = [
            tmp["id"]
            for tmp in response["data"]["projects"]["nodes"]
            if tmp["teams"]["nodes"][0]["id"] == team_id
        ]

        if len(project_ids) > 0:
            logger.warning(
                f"Project {name} already exists in team {team_id}. Skipping creation."
            )
            return project_ids[0]

        query = """
            mutation ProjectCreate($project_name: String!, $team_id: String!) {
                projectCreate(
                    input: {
                        name: $project_name,
                        teamIds: [$team_id]
                        state: "planned"
                    }
                ) {
                    success 
                    project {
                        id
                    }
                }
            }
            """
        variables = {"project_name": name, "team_id": team_id}
        _, response = self._post_request(query, variables)

        return response["data"]["projectCreate"]["project"]["id"]

    def update_project(
        self,
        project_id: str,
        start_date: str = None,
        target_date: str = None,
        state: str = None,
        description: str = None,
        color: str = None,
        sort_order: int = None,
    ) -> str:

        # input validation
        try:
            if start_date is not None:
                t1 = datetime.strptime(start_date, "%Y-%m-%d")
            if target_date is not None:
                t2 = datetime.strptime(target_date, "%Y-%m-%d")

        except ValueError:
            raise Exception("Incorrect date format, should be YYYY-MM-DD")

        if (start_date is not None) and (target_date is not None) and (t1 > t2):
            raise Exception("Start date should be before target date")

        if state not in [
            "backlog",
            "planned",
            "started",
            "paused",
            "completed",
            "canceled",
            None,
        ]:
            raise Exception(
                "State should be one of: backlog, planned, in_progress, paused, completed, canceled"
            )

        # query
        query = """
            mutation UpdateProject($project_id: String!, $startDate: TimelessDate, $targetDate: TimelessDate, $description: String, $state: String = "planned", $color: String, $sortOrder: Float) {
                projectUpdate(
                    id: $project_id
                    input: {
                        startDate: $startDate
                        targetDate: $targetDate
                        description: $description
                        state: $state
                        color: $color
                        sortOrder: $sortOrder
                    }
                ){
                    success
                    project {
                        id
                    }
                }
            }
            """

        # variables
        variables = {"project_id": project_id}
        if start_date is not None:
            variables["startDate"] = start_date
        if target_date is not None:
            variables["targetDate"] = target_date
        if description is not None:
            variables["description"] = description
        if state is not None:
            variables["state"] = state
        if color is not None:
            variables["color"] = color
        if sort_order is not None:
            variables["sort_order"] = sort_order

        # execution
        _, response = self._post_request(query, variables)
        if "errors" in response:
            raise Exception(f"LinearConnector update failed: {response['errors']}")
        return response["data"]["projectUpdate"]["project"]["id"]



import os

import streamlit as st

from asananas import __version__ as VERSION
from asananas.allocation_management import (
    extract_allocation_data,
    visualize_allocation_by_week,
)
from asananas.asana_connector import AsanaConnector
from asananas.asana_linear_bridge import sync_asana_linear

# Page Config
# ###########

st.set_page_config(
    page_title="Asananas Dashboard",
    page_icon="🍍",
    layout="centered",
    initial_sidebar_state="expanded",
)


# Image Paths
# ##########

fp = os.path.dirname(__file__)
FILE_PATH_LOGO = os.path.join(fp, "..", "assets", "logo3.png")
FILE_PATH_ASANA_EXAMPLE_TASK = os.path.join(
    fp, "..", "assets", "asana_example_task.png"
)
FILE_PATH_ASANA_LINEAR_MAPPING = os.path.join(
    fp, "..", "assets", "asana_linear_mapping.png"
)

# Helper Functions
# ################


def _credentials_text_input(variable_name, variable_env_name, password=False):
    variable_value = os.getenv(variable_env_name)

    if variable_value is None:
        variable_value = st.text_input(
            variable_name, type="password" if password else "default"
        )
        if variable_value == "":
            variable_value = None
    else:
        v = "*******" if password else variable_value
        st.success(
            f"{variable_name} found in environment variable {variable_env_name} = {v}"
        )

    return variable_value


def _credentials_select_input(variable_name, variable_env_name, options):

    variable_value = os.getenv(variable_env_name)

    if variable_value is None:
        variable_value = st.selectbox(f"Select {variable_name}", options)
    else:
        st.success(
            f"{variable_name} found in environment variable {variable_env_name} = {variable_value}"
        )

    return variable_value


@st.experimental_memo(ttl=3600)
def _get_workspaces(asana_access_token):
    asana_connector = AsanaConnector(access_token=asana_access_token)
    workspaces = asana_connector.get_workspaces()
    return {w["name"]: w["gid"] for w in workspaces}


@st.experimental_memo(ttl=3600)
def _get_projects(asana_access_token, workspace_id):
    asana_connector = AsanaConnector(access_token=asana_access_token)
    projects = asana_connector.get_projects_for_workspace(workspace_id)
    return {p["name"]: p["gid"] for p in projects}


@st.experimental_memo(ttl=3600)
def _load_data(asana_access_token, asana_project_id):
    return AsanaConnector(access_token=asana_access_token).get_all_tasks_for_project(
        asana_project_id
    )


# Header and Credentials
# ######################

st.header("Asananas Dashboard")

st.markdown(
    """
    Asananas helps you with your project management in [Asana](http://asana.com/). It assumes that a single Asana project, e.g. called "Company Workstreams", is used for high-level project planning of your company or your team. Different projects/workstreams are reflected in Asana as tasks in the dedicated Asana project. These Asana tasks can be assigned to different people and have a start and a due date.

    Asananas helps you to visualize the allocation of your team members over time. Additionally, it helps you to sync your Asana tasks with [Linear](http://linear.app/), a project management tool for software development teams. This is particularity useful for projects that are more technical in nature.
    """
)

with st.expander("Settings", expanded=False):

    st.markdown(
        "**ASANA**: To connect this dashboard to your Asana workspace, you need to provide your Asana access token and select a workspace and project. You can get your token [here](https://app.asana.com/0/my-apps)."
    )

    asana_access_token = _credentials_text_input(
        "Asana Access Token", "ASANA_ACCESS_TOKEN", password=True
    )

    if asana_access_token is not None:

        workspaces = _get_workspaces(asana_access_token)
        asana_workspace_name = _credentials_select_input(
            "Asana Workspace Name", "ASANA_WORKSPACE_NAME", list(workspaces.keys())
        )
        asana_workspace_id = workspaces[asana_workspace_name]

        if asana_workspace_id is not None:
            projects = _get_projects(asana_access_token, asana_workspace_id)
            asana_project_name = _credentials_select_input(
                "Asana Project Name", "ASANA_PROJECT_NAME", list(projects.keys())
            )
            asana_project_id = projects[asana_project_name]

    else:
        st.error("Provide a valid asana access token to continue.")
        asana_project_id = None

    st.markdown("---")

    st.markdown(
        "**LINEAR**: If you additionally want to use the Linear bridge you need to provide a Linear Access token as well as a team name."
    )

    linear_access_token = _credentials_text_input(
        "Linear Access Token", "LINEAR_ACCESS_TOKEN", password=True
    )
    linear_team_name = _credentials_text_input("Liner Team Name", "LINEAR_TEAM_NAME")

    st.markdown("---")

    st.markdown(
        "For convenience, you can provide all this information also via environment variables called `ASANA_ACCESS_TOKEN`, `ASANA_WORKSPACE_NAME`, `ASANA_PROJECT_NAME`, `LINEAR_ACCESS_TOKEN`, `LINEAR_TEAM_NAME`. This is useful in case this dashboard should be deployed for one specific team."
    )


with st.expander("Setup", expanded=False):
    st.markdown(
        """
    **Resource Allocation Management**: With a growing number of tasks the  resource allocation becomes harder and harder to oversee and it gets even more complicated when multiple people work on a single task. To avoid this problem, Asananas provides a simple way to visualize the resource allocation of your Asana project.
    To do so, you need to specify the required resources for each task in the task description by adding a custom field called  'Allocation'. The required resources are specified in the following format:
    - `Name1: 2d`: You can either specify the required resources in days. In this example, the task requires 2 days of Name1's time.
    - `Name1: 30%`: You can also specify the required resources in percent. In this example, the task requires 30% of the Name1's work time.
    - `Name1: 3d, Name2: 70%`: If you want to specify multiple resources, you can separate them by a comma.
    The image below shows an example task with the required resources specified in the Allocation field.
    """
    )
    _, c, _ = st.columns([1, 2, 1])
    with c:
        st.image(FILE_PATH_ASANA_EXAMPLE_TASK, caption="Asana Example Task")
    st.markdown(
        """
    ---

    **Linear Bridge**: From some tasks in your Asana project it might be useful to track the progress in Linear. Here we map a project in Asana to a team in Linear and a task in Asana to projects in Linear. See the image below for an illustration. To sync the start and due dates of the Asana task with the Linear project, provide your Linear API key and the Linear team name in the settings above and press the button below. All Asana tasks with the tag `Linear Project` will be synced (see screenshot above).
    """
    )
    _, c, _ = st.columns([1, 4, 1])
    with c:
        st.image(FILE_PATH_ASANA_LINEAR_MAPPING, caption="Mapping from Asana to Linear")


# Sidebar Logo
# ###########r

st.sidebar.image(FILE_PATH_LOGO, use_column_width=True)
st.sidebar.markdown(
    f"<div style='text-align: center;'>v{VERSION}</div>", unsafe_allow_html=True
)
st.sidebar.markdown("---")


# Load Data
# #########
df_asana_tasks = _load_data(asana_access_token, asana_project_id)

st.sidebar.markdown(
    "To speed up the dashboard, the data is cached for one hour. If you updated your Asana tasks press the button below to reload the data."
)

if st.sidebar.button("Clear Cache & Rerun"):
    _get_workspaces.clear()
    _get_projects.clear()
    _load_data.clear()
    st.experimental_rerun()


# Resource Allocation
# ###################

st.markdown("---")

st.subheader("⌛ Resource Allocation")

(
    df_allocation_data,
    projects_with_no_allocation,
    projects_with_broken_allocation,
) = extract_allocation_data(df_asana_tasks, n_workdays_per_week=5)

if len(df_allocation_data) == 0:
    st.warning(
        f"Could not find any resource allocation data for the Asana project {asana_project_name}."
    )
else:
    st.markdown(
        "Individual tasks specified in your Asana project require different human resources. By specifying the required resources in the Asana task, you can visualize the resource allocation of your team. The following image shows the resource allocation by week:"
    )

    # filter by name
    names = list(df_allocation_data.name.unique())
    selected_names = st.multiselect("Filter by names", names, default=names)
    df_allocation_data_plot = df_allocation_data[
        df_allocation_data["name"].isin(selected_names)
    ]

    # visualize
    fig = visualize_allocation_by_week(df_allocation_data_plot)
    st.plotly_chart(fig)

    # warnings and errors
    if len(projects_with_no_allocation) > 0:
        st.warning(
            f"The following {len(projects_with_no_allocation)} projects have no resource allocated: {', '.join(projects_with_no_allocation)}"
        )
    if len(projects_with_broken_allocation) > 0:
        st.error(
            f"The following {len(projects_with_broken_allocation)} projects have an allocation that could not be decoded: {', '.join(projects_with_broken_allocation)}"
        )


# Linear Bridge
# #############

st.markdown("---")

st.subheader("🔄 Linear Bridge")

c1, c2, c3 = st.columns(3)
with c1:
    auto_create_linear_projects = st.checkbox("Auto create Linear projects", value=True)
with c2:
    sync_projects = st.checkbox("Sync project timelines", value=True)
with c3:
    cancel_linear_projects = st.checkbox("Auto cancel Linear projects", value=True)

if (
    asana_project_id is not None
    and linear_access_token is not None
    and linear_team_name is not None
):
    if st.button("Run Asana-Linear Bridge"):
        with st.spinner("Sync in progress..."):

            sync_asana_linear(
                asana_project_id,
                asana_access_token,
                linear_team_name,
                linear_access_token,
                auto_create_linear_projects,
                sync_projects,
                cancel_linear_projects,
            )
else:
    st.warning(
        "Make sure you provided the linear access token, linear team name and linear access token in the settings above."
    )



import os
import typing as t

import asana
import pandas as pd


class AsanaConnector:
    def __init__(self, access_token: str = None):

        if access_token is None:
            access_token = os.getenv("ASANA_ACCESS_TOKEN")

        if access_token is None:
            raise Exception("No Asana access token provided")

        self.client = asana.Client.access_token(access_token)

    def get_workspaces(self):
        return list(self.client.workspaces.find_all())

    def get_projects_for_workspace(self, workspace_id):
        return list(self.client.projects.find_by_workspace(workspace_id))

    def _find_section_by_project_id(self, task_info, project_id):
        for membership in task_info["memberships"]:
            if membership["project"]["gid"] == project_id:
                if "section" in membership:
                    return membership["section"]["name"]
        return None

    def _find_allocation(self, task_info):
        for custom_field in task_info["custom_fields"]:
            if custom_field["name"] == "Allocation":
                return custom_field["text_value"]
        return None

    def get_all_tasks_for_project(self, project_id) -> t.List[t.Dict]:
        task_infos = []
        for task in self.client.tasks.find_by_project(project_id):
            task_info = self.client.tasks.get_task(task["gid"])

            if task_info["resource_subtype"] != "default_task":
                continue

            assignee = (
                None if task_info["assignee"] is None else task_info["assignee"]["name"]
            )
            tags = [t["name"] for t in task_info["tags"]]

            task_infos.append(
                {
                    "asana_task_id": task_info["gid"],
                    "asana_task_name": task_info["name"],
                    "asana_start_on": task_info["start_on"],
                    "asana_due_on": task_info["due_on"],
                    "asana_completed": task_info["completed"],
                    "asana_assignee": assignee,
                    "asana_url": task_info["permalink_url"],
                    "asana_linear_project": "Linear Project" in tags,
                    "asana_section": self._find_section_by_project_id(
                        task_info, project_id
                    ),
                    "asana_allocation": self._find_allocation(task_info),
                }
            )

        if len(task_infos) == 0:
            columns = [
                "asana_task_id",
                "asana_task_name",
                "asana_start_on",
                "asana_due_on",
                "asana_completed",
                "asana_assignee",
                "asana_url",
                "asana_linear_project",
                "asana_section",
                "asana_allocation",
            ]
            return pd.DataFrame(columns=columns)

        return pd.DataFrame(task_infos)



import sys
from os import path as p

from streamlit.web import cli as stcli


def run():
    dashboard_file = p.join(p.abspath(p.dirname(__file__)), "main.py")
    sys.argv = ["streamlit", "run", dashboard_file]
    sys.exit(stcli.main())



import typing as t
from datetime import datetime

import pandas as pd
from loguru import logger

from asananas.asana_connector import AsanaConnector
from asananas.linear_connector import LinearConnector

COLORS = [
    "#1F77B4",
    "#FF7F0E",
    "#2CA02C",
    "#D62728",
    "#9467BD",
    "#8C564B",
    "#CFECF9",
    "#7F7F7F",
    "#BCBD22",
    "#17BECF",
]


def sync_asana_linear(
    asana_project_id,
    asana_access_token,
    linear_team_name,
    linear_access_token,
    auto_create_linear_projects,
    sync_projects,
    cancel_linear_projects,
):

    logger.info("Fetching data from Asana")
    df_asana_tasks = AsanaConnector(
        access_token=asana_access_token
    ).get_all_tasks_for_project(asana_project_id)
    df_asana_tasks = df_asana_tasks[df_asana_tasks.asana_linear_project]

    logger.info("Fetching data from Linear")
    linear_connector = LinearConnector(access_token=linear_access_token)
    team_id, _, df_linear_projects = linear_connector.get_team_and_projects(
        linear_team_name
    )

    logger.info("Merging data")
    df = df_asana_tasks.merge(
        df_linear_projects,
        left_on="asana_task_name",
        right_on="linear_project_name",
        how="outer",
    )
    df["exists_on_asana"] = ~pd.isna(df.asana_task_id)
    df["exists_on_linear"] = ~pd.isna(df.linear_project_id)
    df["ongoing"] = pd.to_datetime(df.asana_start_on) < datetime.now()

    df_color_mapping = pd.DataFrame(
        [
            {"asana_section": s, "linear_color": COLORS[i % len(COLORS)]}
            for i, s in enumerate(df.asana_section.unique())
        ]
    )
    df = df.merge(df_color_mapping, on="asana_section", how="left")

    df = df.sort_values(by=["asana_section", "asana_start_on"]).reset_index(drop=True)

    for index, row in df.iterrows():

        if not row.asana_linear_project:
            continue

        # create linear project
        if (
            auto_create_linear_projects
            and row.exists_on_asana
            and not row.exists_on_linear
        ):

            row.linear_project_id = linear_connector.create_project(
                name=row["asana_task_name"],
                team_id=team_id,
            )

            row.exists_on_linear = True
            logger.info(
                f"Created Linear project {row.linear_project_id} for Asana task {row['asana_task_name']}"
            )

        # sync linear project
        if sync_projects & row.exists_on_asana & row.exists_on_linear:

            state = "planned"

            if row.ongoing:
                state = "started"

            if row.asana_completed:
                state = "completed"

            linear_connector.update_project(
                project_id=row.linear_project_id,
                start_date=row.asana_start_on,
                target_date=row.asana_due_on,
                state=state,
                description=f"Asana section: {row.asana_section}/nAsana URL: {row.asana_url}",
                color=row.linear_color,
                sort_order=index,
            )

            logger.info(
                f"Updated Linear project {row.linear_project_id} according to Asana task {row['asana_task_name']}"
            )

        # cancel linear project
        if cancel_linear_projects and not row.exists_on_asana and row.exists_on_linear:
            if row.linear_state != "cancelled":
                linear_connector.update_project(
                    project_id=row.linear_project_id, state="canceled"
                )
                logger.info(
                    f"Cancelled Linear project '{row.linear_project_name}' because no asana task exists."
                )



import re
from datetime import datetime, timedelta

import pandas as pd
import plotly.express as px


def _get_all_work_days(t1, t2, n_workdays_per_week):
    dates = []
    for i in range((t2 - t1).days + 1):
        date = t1 + timedelta(days=i)
        if date.isoweekday() <= n_workdays_per_week:
            dates.append(date)
    return dates


def extract_allocation_data(df_asana_tasks, n_workdays_per_week=5):

    projects_with_no_allocation = []
    projects_with_broken_allocation = []

    data = []

    for _, row in df_asana_tasks.iterrows():

        # skip tasks with no allocation
        if row.asana_allocation is None:
            projects_with_no_allocation.append(row.asana_task_name)
            continue

        # decode allocation
        pattern = r"([A-Z]+): ([0-9]+)(%|d)"
        allocations = re.findall(pattern, row.asana_allocation)
        if len(allocations) == 0:
            projects_with_broken_allocation.append(row.asana_task_name)
            continue

        # convert dates
        t1 = datetime.strptime(row.asana_start_on, "%Y-%m-%d")
        t2 = datetime.strptime(row.asana_due_on, "%Y-%m-%d")
        dates = _get_all_work_days(t1, t2, n_workdays_per_week=n_workdays_per_week)

        # build up data
        for date in dates:

            for allocation_info in allocations:

                allocation_value = float(allocation_info[1])
                allocation_unit = allocation_info[2]

                if allocation_unit == "%":
                    allocation = allocation_value / 100
                elif allocation_unit == "d":
                    allocation = allocation_value / len(dates)

                data.append(
                    {
                        "date": date,
                        "name": allocation_info[0],
                        "allocation": allocation,
                        "project": row.asana_task_name,
                    }
                )

    return (
        pd.DataFrame(data),
        projects_with_no_allocation,
        projects_with_broken_allocation,
    )


def visualize_allocation_by_week(df_allocation_data, n_workdays_per_week=5):

    # create the week field
    df_allocation_data["week"] = df_allocation_data["date"].apply(
        lambda x: f"{x.isocalendar().year}-CW{x.isocalendar().week}"
    )

    # group by name, project and week
    df_tmp = (
        df_allocation_data.groupby(["name", "project", "week"])
        .aggregate(
            allocation=("allocation", "first"),
            n_days_in_this_week=("date", "count"),
            week_of=("date", "min"),
        )
        .reset_index()
        .sort_values("week")
    )
    df_tmp["week_of"] = df_tmp.week_of.apply(lambda date: date.strftime("%Y-%m-%d"))
    df_tmp["allocation"] = (
        df_tmp.allocation * df_tmp.n_days_in_this_week / n_workdays_per_week
    )

    # basic bar plot
    fig = px.bar(
        df_tmp,
        x="week",
        y="allocation",
        color="project",
        title="Resource Allocation by Week",
        facet_row="name",
        hover_data=["week_of"],
    )

    # maximum allocation marker
    fig.add_hline(y=1)

    # grey out the past
    t = datetime.now()
    x1 = f"{t.year}-CW{t.isocalendar().week}"
    fig.add_vrect(
        x0=df_tmp.week.min(), x1=x1, line_width=0, fillcolor="black", opacity=0.4
    )

    # update the size of the plot
    fig.update_layout(
        autosize=False,
        width=1000,
        height=df_tmp.name.nunique() * 200,
    )

    # update the x-range
    t = datetime.now() - timedelta(days=7 * 4)
    x0 = f"{t.year}-CW{t.isocalendar().week}"
    idx_1 = (df_tmp.week.drop_duplicates().sort_values() == x0).argmax()

    idx_2 = (df_tmp.week.drop_duplicates().sort_values() == df_tmp.week.max()).argmax()
    fig.update_xaxes(range=[idx_1, idx_2])
    fig.update_yaxes(range=[0, 1.5])

    return fig
